import pandas as pd
from google.colab import auth

auth.authenticate_user()

import gspread
from google.auth import default

creds, _ = default()

gc = gspread.authorize(creds)


def read_google_spreadsheet(spreadsheet_filename: str):
    """Input spreadsheet name. Return as gc spreadsheet"""
    return gc.open(spreadsheet_filename)


def list_all_worksheets(spreadsheet_filename: str):
    """Input spreadsheet name. Return list of all worksheet titles"""
    sh = read_google_spreadsheet(spreadsheet_filename)
    worksheet_objs = sh.worksheets()
    return [wksheet.title for wksheet in worksheet_objs]


def read_google_worksheet(spreadsheet_filename: str, worksheet_name: str):
    """Input name of spreadsheet and worksheet name. Return google worksheet"""
    sh = read_google_spreadsheet(spreadsheet_filename)
    gc_worksheet = sh.worksheet(worksheet_name)
    return gc_worksheet


#### main functions we will probably use
def worksheet_to_dataframe(spreadsheet_filename: str, worksheet_name: str):
    """Input spreadsheet and worksheet names. Transform and return as dataframe"""
    worksheet = read_google_worksheet(spreadsheet_filename, worksheet_name)
    values_list = worksheet.get_all_values()
    df = pd.DataFrame(values_list).fillna("")
    df.columns = df.iloc[0]
    return df.drop(df.index[0])


def update_google_worksheet_from_df(
    worksheet: gspread.models.Worksheet, df: pd.DataFrame
):
    """Input google worksheet and pandas dataframe. Update current worksheet with dataframe
    contents and save directly to the file"""
    return worksheet.update([df.columns.values.tolist()] + df.values.tolist())



import math
from mlcm import mlcm
import pandas as pd


def one_hot_encoding(data, labels_in_order):
    one_hot_encode = []
    for line in data:
        temp = []
        for label in labels_in_order:
            if label in line:
                temp.append(1)
            else:
                temp.append(0)
        one_hot_encode.append(temp)
    return one_hot_encode


def count_error_types(correct, predict, labels_in_order):
    label_error_map = {}
    for x, label in enumerate(labels_in_order):
        temp = {"TP": 0, "TN": 0, "FP": 0, "FN": 0}
        for (cor, pre) in zip(correct, predict):
            if cor[x] == pre[x]:
                if cor[x]:
                    temp["TP"] += 1
                else:
                    temp["TN"] += 1
            else:
                if cor[x]:
                    temp["FN"] += 1
                else:
                    temp["FP"] += 1

        label_error_map[label] = temp
    return label_error_map


def calculate_supports(true_labels_encoded, labels_in_order):
    supports_count = {key: 0 for key in labels_in_order}
    for item in true_labels_encoded:
        for (label_name, label) in zip(labels_in_order, item):
            if label:
                supports_count[label_name] = supports_count[label_name] + 1

    return supports_count


def calculate_metrics(
    label_error_map, supports_count
):  # include average type e.g. micro/macro/binary/weighted etc
    def accuracy(TP, TN, FP, FN):
        try:
            return (TP + TN) / (TP + TN + FP + FN)
        except:
            return 0

    def precision(TP, TN, FP, FN):
        try:
            return TP / (TP + FP)
        except:
            return 0

    def recall(TP, TN, FP, FN):
        try:
            return TP / (TP + FN)
        except:
            return 0

    def f1(precision, recall):
        try:
            return 2 * (precision * recall) / (precision + recall)
        except:
            return 0

    def f0_5(precision, recall):
        try:
            return (1.25 * precision * recall) / (0.25 * (precision + recall))
        except:
            return 0

    result = {}
    for label in label_error_map.keys():
        matrics_map = {"accuracy": 0, "precision": 0, "recall": 0, "f1": 0, "f0.5": 0}
        error_map = label_error_map[label]

        for error in matrics_map.keys():
            if error == "accuracy":
                matrics_map["accuracy"] = accuracy(
                    error_map["TP"], error_map["TN"], error_map["FP"], error_map["FN"]
                )
            if error == "precision":
                matrics_map["precision"] = precision(
                    error_map["TP"], error_map["TN"], error_map["FP"], error_map["FN"]
                )
            if error == "recall":
                matrics_map["recall"] = recall(
                    error_map["TP"], error_map["TN"], error_map["FP"], error_map["FN"]
                )
            if error == "f1":
                matrics_map["f1"] = f1(matrics_map["precision"], matrics_map["recall"])
            if error == "f0.5":
                matrics_map["f0.5"] = f1(
                    matrics_map["precision"], matrics_map["recall"]
                )

        # for error in matrics_map.keys():
        #     matrics_map[error] = math.floor(matrics_map[error] * 1000) / 1000

        result[label] = matrics_map

    average_map = {"accuracy": [], "precision": [], "recall": [], "f1": [], "f0.5": []}
    for key in result.keys():
        for type_ in average_map.keys():
            average_map[type_].append(result[key][type_])
    for type_ in average_map.keys():
        global_scores_ = sum(average_map[type_]) / len(average_map[type_])
        # global_scores_ = math.floor(global_scores_ * 1000) / 1000
        average_map[type_] = global_scores_
    result["average"] = average_map

    weighted_map = {"accuracy": [], "precision": [], "recall": [], "f1": [], "f0.5": []}
    for type_ in matrics_map.keys():
        scores = []
        for label in label_error_map.keys():
            scores.append(result[label][type_])
        global_scores_ = sum(
            [
                (score * support)
                for (score, support) in zip(scores, supports_count.values())
            ]
        ) / sum(supports_count.values())
        # global_scores_ = math.floor(global_scores_ * 1000) / 1000
        weighted_map[type_] = global_scores_
    result["weighted_average"] = weighted_map

    return result


def get_conf_matrix(correct, predict, labels_in_order, normalized=False):
    conf_mat, normalized_conf_mat = mlcm.cm(correct, predict, False)
    if normalized:
        conf_mat = normalized_conf_mat
    conf_matrix_df = pd.DataFrame(
        conf_mat, columns=labels_in_order + ["NLP"], index=labels_in_order + ["NTL"]
    )
    return conf_matrix_df


def print_classification_report(correct, predict):
    conf_mat, _ = mlcm.cm(correct, predict, False)
    one_vs_rest = mlcm.stats(conf_mat, False)
    return one_vs_rest



from grpc_requests import Client

company_pages_grpc_endpoint = "company-pages.grpc.delphai.com:80"


class ScrapedCompanyPage:
    def __init__(self):
        client = Client(company_pages_grpc_endpoint)

        health = client.service("grpc.health.v1.Health")
        assert health.method_names == ("Check", "Watch")

        result = health.Check()
        assert result == {"status": "SERVING"}

        self.CompanyPages = client.service("delphai.CompanyPages")

    def list_domains(self, prefix: str, limit: int = 20):
        stream = self.CompanyPages.list_domains({"prefix": prefix, "limit": limit})
        return stream

    def list_pages(self, domain: str):
        stream = self.CompanyPages.list_urls({"domain": domain})
        return stream

    def get_content_for_class(
        self,
        content_type: str,
        domain: str = None,
        company_id: str = None,
        page_classes: list = None,
        classifier: str = None,
        language: str = None,
    ):
        request = {}
        request["type"] = content_type

        if domain or company_id:
            if domain:
                request["domain"] = domain
            if company_id:
                request["company_id"] = company_id
        else:
            raise ValueError("Should provide a domain or a company_id.")

        if page_classes:
            request["page_classes"] = page_classes
        if classifier:
            request["classifier"] = classifier
        if language:
            request["language"] = language

        stream = self.CompanyPages.pages(request)
        page_list = list(stream)
        return page_list, len(page_list)

    def get_page_metadata(self, url: str):
        stream = self.CompanyPages.get_page_metadata({"url": url})
        return stream

    def get_page_content(self, url: str, content_type: str, language: str = None):
        request = {}
        request["url"] = url
        request["type"] = content_type
        if language:
            request["language"] = language

        stream = self.CompanyPages.page(request)
        return stream



# Following https://huggingface.co/docs/transformers/serialization

import os
from transformers import AutoTokenizer


def make_onnx(
    path_to_model: str,
    pretrained_model_name: str,
    feature: str = "sequence-classification",
    atol: float = 0,
) -> None:
    tokenizer = AutoTokenizer.from_pretrained(
        pretrained_model_name_or_path=pretrained_model_name
    )
    tokenizer.save_pretrained(path_to_model)
    command = f"python -m transformers.onnx --model={path_to_model} {path_to_model}"
    if feature:
        command += f" --feature {feature}"
    if atol:
        command += f" --atol {atol}"
    os.system(command)



from tqdm.notebook import tqdm
import torch
from torch.utils.data import Dataset, DataLoader
import numpy as np
from torch.nn.functional import cross_entropy


class EarlyStopping(object):
    def __init__(self, mode="min", min_delta=0, patience=10, percentage=False):
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta, percentage)

        if patience == 0:
            self.is_better = lambda a, b: True
            self.step = lambda a: False

    def step(self, metrics):
        if self.best is None:
            self.best = metrics
            return False

        if np.isnan(metrics):
            return True

        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.best = metrics
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            print("terminating because of early stopping!")
            return True

        return False

    def _init_is_better(self, mode, min_delta, percentage):
        if mode not in {"min", "max"}:
            raise ValueError("mode " + mode + " is unknown!")
        if not percentage:
            if mode == "min":
                self.is_better = lambda a, best: a < best - min_delta
            if mode == "max":
                self.is_better = lambda a, best: a > best + min_delta
        else:
            if mode == "min":
                self.is_better = lambda a, best: a < best - (best * min_delta / 100)
            if mode == "max":
                self.is_better = lambda a, best: a > best + (best * min_delta / 100)


class Dataset(Dataset):
    r"""PyTorch Dataset class for loading data.

    This is where the data parsing happens and where the text gets encoded using
    loaded tokenizer.

    This class is built with reusability in mind: it can be used as is as long
      as the `dataloader` outputs a batch in dictionary format that can be passed
      straight into the model - `model(**batch)`.

    Arguments:

      path (:obj:`str`):
          Path to the data partition.

      use_tokenizer (:obj:`transformers.tokenization_?`):
          Transformer type tokenizer used to process raw text into numbers.

      labels_ids (:obj:`dict`):
          Dictionary to encode any labels names into numbers. Keys map to
          labels names and Values map to number associated to those labels.

      max_sequence_len (:obj:`int`, `optional`)
          Value to indicate the maximum desired sequence to truncate or pad text
          sequences. If no value is passed it will used maximum sequence size
          supported by the tokenizer and model.

    """

    def __init__(
        self,
        df,
        content_column,
        use_tokenizer,
        labels_ids=None,
        label_column=None,
        max_sequence_len=None,
        for_pred=False,
    ):

        # Check max sequence length.
        max_sequence_len = (
            use_tokenizer.max_len if max_sequence_len is None else max_sequence_len
        )
        texts = []
        if not for_pred:
            labels = []
            label_count = {}
            # Since the labels are defined by folders with data we loop
            # through each label.
            print("Reading rows...")
            for index, row in tqdm(df.iterrows(), total=len(df)):
                if row[label_column] not in label_count:
                    label_count[row[label_column]] = 0
                label_count[row[label_column]] += 1
                texts.append(row[content_column])
                labels.append(labels_ids[row[label_column]])
            print(f"Total data points map: {label_count}")

        else:
            texts = list(df[content_column].values)

        # Number of exmaples.
        self.n_examples = len(texts)
        # Use tokenizer on texts. This can take a while.
        print("Using tokenizer on all texts. This can take a while...")
        self.inputs = use_tokenizer(
            texts,
            add_special_tokens=True,
            truncation=True,
            padding=True,
            return_tensors="pt",
            max_length=max_sequence_len,
        )
        # Get maximum sequence length.
        self.sequence_len = self.inputs["input_ids"].shape[-1]
        print("Texts padded or truncated to %d length!" % self.sequence_len)
        if not for_pred:
            # Add labels.
            self.inputs.update({"labels": torch.tensor(labels)})

        print("Finished!\n")

        return

    def __len__(self):
        r"""When used `len` return the number of examples."""

        return self.n_examples

    def __getitem__(self, item):
        r"""Given an index return an example from the position.

        Arguments:

          item (:obj:`int`):
              Index position to pick an example to return.

        Returns:
          :obj:`Dict[str, object]`: Dictionary of inputs that feed into the model.
          It holddes the statement `model(**Returned Dictionary)`.

        """

        return {key: self.inputs[key][item] for key in self.inputs.keys()}


def train(dataloader, model_, optimizer_, scheduler_, device_):
    r"""
    Train pytorch model on a single pass through the data loader.

    It will use the global variable `model` which is the transformer model
    loaded on `_device` that we want to train on.

    This function is built with reusability in mind: it can be used as is as long
      as the `dataloader` outputs a batch in dictionary format that can be passed
      straight into the model - `model(**batch)`.

    Arguments:

        dataloader (:obj:`torch.utils.data.dataloader.DataLoader`):
            Parsed data into batches of tensors.

        optimizer_ (:obj:`transformers.optimization.AdamW`):
            Optimizer used for training.

        scheduler_ (:obj:`torch.optim.lr_scheduler.LambdaLR`):
            PyTorch scheduler.

        device_ (:obj:`torch.device`):
            Device used to load tensors before feeding to model.

    Returns:

        :obj:`List[List[int], List[int], float]`: List of [True Labels, Predicted
          Labels, Train Average Loss].
    """

    # Tracking variables.
    predictions_labels = []
    true_labels = []
    # Total loss for this epoch.
    total_loss = 0

    # Put the model into training mode.
    model_.train()

    # For each batch of training data...
    for batch in tqdm(dataloader, total=len(dataloader)):

        # Add original labels - use later for evaluation.
        true_labels += batch["labels"].numpy().flatten().tolist()

        # move batch to device
        batch = {k: v.type(torch.long).to(device_) for k, v in batch.items()}

        # Always clear any previously calculated gradients before performing a
        # backward pass.
        model_.zero_grad()

        # Perform a forward pass (evaluate the model on this training batch).
        # This will return the loss (rather than the model output) because we
        # have provided the `labels`.
        # The documentation for this a bert model function is here:
        # https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification
        outputs = model_(**batch)

        # The call to `model` always returns a tuple, so we need to pull the
        # loss value out of the tuple along with the logits. We will use logits
        # later to calculate training accuracy.
        loss, logits = outputs[:2]

        # Accumulate the training loss over all of the batches so that we can
        # calculate the average loss at the end. `loss` is a Tensor containing a
        # single value; the `.item()` function just returns the Python value
        # from the tensor.
        total_loss += loss.item()

        # Perform a backward pass to calculate the gradients.
        loss.backward()

        # Clip the norm of the gradients to 1.0.
        # This is to help prevent the "exploding gradients" problem.
        torch.nn.utils.clip_grad_norm_(model_.parameters(), 1.0)

        # Update parameters and take a step using the computed gradient.
        # The optimizer dictates the "update rule"--how the parameters are
        # modified based on their gradients, the learning rate, etc.
        optimizer_.step()

        # Update the learning rate.
        scheduler_.step()

        # Move logits and labels to CPU
        logits = logits.detach().cpu().numpy()

        # Convert these logits to list of predicted labels values.
        predictions_labels += logits.argmax(axis=-1).flatten().tolist()

    # Calculate the average loss over the training data.
    avg_epoch_loss = total_loss / len(dataloader)

    # Return all true labels and prediction for future evaluations.
    return true_labels, predictions_labels, avg_epoch_loss


def validation(dataloader, model_, device_):
    r"""Validation function to evaluate model performance on a
    separate set of data.

    This function will return the true and predicted labels so we can use later
    to evaluate the model's performance.

    This function is built with reusability in mind: it can be used as is as long
      as the `dataloader` outputs a batch in dictionary format that can be passed
      straight into the model - `model(**batch)`.

    Arguments:

      dataloader (:obj:`torch.utils.data.dataloader.DataLoader`):
            Parsed data into batches of tensors.

      device_ (:obj:`torch.device`):
            Device used to load tensors before feeding to model.

    Returns:

      :obj:`List[List[int], List[int], float]`: List of [True Labels, Predicted
          Labels, Train Average Loss]
    """

    # Tracking variables
    predictions_labels = []
    true_labels = []
    # total loss for this epoch.
    total_loss = 0

    # Put the model in evaluation mode--the dropout layers behave differently
    # during evaluation.
    model_.eval()

    # Evaluate data for one epoch
    for batch in tqdm(dataloader, total=len(dataloader)):

        # add original labels
        true_labels += batch["labels"].numpy().flatten().tolist()

        # move batch to device
        batch = {k: v.type(torch.long).to(device_) for k, v in batch.items()}

        # Telling the model not to compute or store gradients, saving memory and
        # speeding up validation
        with torch.no_grad():

            # Forward pass, calculate logit predictions.
            # This will return the logits rather than the loss because we have
            # not provided labels.
            # token_type_ids is the same as the "segment ids", which
            # differentiates sentence 1 and 2 in 2-sentence tasks.
            # The documentation for this `model` function is here:
            # https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification
            outputs = model_(**batch)

            # The call to `model` always returns a tuple, so we need to pull the
            # loss value out of the tuple along with the logits. We will use logits
            # later to to calculate training accuracy.
            loss, logits = outputs[:2]

            # Move logits and labels to CPU
            logits = logits.detach().cpu().numpy()

            # Accumulate the training loss over all of the batches so that we can
            # calculate the average loss at the end. `loss` is a Tensor containing a
            # single value; the `.item()` function just returns the Python value
            # from the tensor.
            total_loss += loss.item()

            # get predicitons to list
            predict_content = logits.argmax(axis=-1).flatten().tolist()

            # update list
            predictions_labels += predict_content

    # Calculate the average loss over the training data.
    avg_epoch_loss = total_loss / len(dataloader)

    # Return all true labels and prediciton for future evaluations.
    return true_labels, predictions_labels, avg_epoch_loss


def prediction(dataloader, model_, device_):
    r"""Prediction function to evaluate model performance on a
    separate set of data.

    This function will return the true and predicted labels so we can use later
    to evaluate the model's performance.

    This function is built with reusability in mind: it can be used as is as long
      as the `dataloader` outputs a batch in dictionary format that can be passed
      straight into the model - `model(**batch)`.

    Arguments:

      dataloader (:obj:`torch.utils.data.dataloader.DataLoader`):
            Parsed data into batches of tensors.

      device_ (:obj:`torch.device`):
            Device used to load tensors before feeding to model.

    Returns:

      :obj:`List[List[int], List[int], float]`: List of [True Labels, Predicted
          Labels, Train Average Loss]
    """

    # Tracking variables
    predictions_labels = []
    predictions_probabilities = []
    #     true_labels = []
    # total loss for this epoch.
    total_loss = 0

    # Put the model in evaluation mode--the dropout layers behave differently
    # during evaluation.
    model_.eval()

    # Evaluate data for one epoch
    for batch in tqdm(dataloader, total=len(dataloader)):

        #         # add original labels
        #         true_labels += batch['labels'].numpy().flatten().tolist()

        # move batch to device
        batch = {k: v.type(torch.long).to(device_) for k, v in batch.items()}

        # Telling the model not to compute or store gradients, saving memory and
        # speeding up validation
        with torch.no_grad():

            # Forward pass, calculate logit predictions.
            # This will return the logits rather than the loss because we have
            # not provided labels.
            # token_type_ids is the same as the "segment ids", which
            # differentiates sentence 1 and 2 in 2-sentence tasks.
            # The documentation for this `model` function is here:
            # https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification
            outputs = model_(**batch)

            # The call to `model` always returns a tuple, so we need to pull the
            # loss value out of the tuple along with the logits. We will use logits
            # later to to calculate training accuracy.
            logits = outputs[0]
            probs = torch.sigmoid(logits)

            # Move logits and labels to CPU
            logits = logits.detach().cpu().numpy()
            probs = probs.detach().cpu().numpy()

            # get predicitons to list
            predict_content = logits.argmax(axis=-1).flatten().tolist()

            # update list
            predictions_labels += predict_content
            predictions_probabilities.append(probs.tolist())

    predictions_probabilities = [
        item for sublist in predictions_probabilities for item in sublist
    ]
    # Return all true labels and prediciton for future evaluations.
    return predictions_labels, predictions_probabilities


def error_analysis(dataloader, model_, device_):
    # Tracking variables
    predictions_labels = []
    true_labels = []
    # total loss for this epoch.
    total_loss = 0
    all_loss = []

    # Put the model in evaluation mode--the dropout layers behave differently
    # during evaluation.
    model_.eval()

    # Evaluate data for one epoch
    for batch in tqdm(dataloader, total=len(dataloader)):

        # add original labels
        true_labels += batch["labels"].numpy().flatten().tolist()

        # move batch to device
        batch = {k: v.type(torch.long).to(device_) for k, v in batch.items()}

        # Telling the model not to compute or store gradients, saving memory and
        # speeding up validation
        with torch.no_grad():

            # Forward pass, calculate logit predictions.
            # This will return the logits rather than the loss because we have
            # not provided labels.
            # token_type_ids is the same as the "segment ids", which
            # differentiates sentence 1 and 2 in 2-sentence tasks.
            # The documentation for this `model` function is here:
            # https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification
            outputs = model_(**batch)

            # The call to `model` always returns a tuple, so we need to pull the
            # loss value out of the tuple along with the logits. We will use logits
            # later to to calculate training accuracy.
            loss, logits = outputs[:2]

            loss_ = cross_entropy(logits, batch["labels"], reduction="none")
            all_loss.extend(loss_.cpu().numpy())

            # Move logits and labels to CPU
            logits = logits.detach().cpu().numpy()

            # Accumulate the training loss over all of the batches so that we can
            # calculate the average loss at the end. `loss` is a Tensor containing a
            # single value; the `.item()` function just returns the Python value
            # from the tensor.
            total_loss += loss.item()

            # get predicitons to list
            predict_content = logits.argmax(axis=-1).flatten().tolist()

            # update list
            predictions_labels += predict_content

    # Calculate the average loss over the training data.
    avg_epoch_loss = total_loss / len(dataloader)

    # Return all true labels and prediciton for future evaluations.
    return true_labels, predictions_labels, avg_epoch_loss, all_loss



from trafilatura import extract
import trafilatura
from goose3 import Goose
import re
import justext
import logging
from delphai_ml_utils.text_processing import remove_emoji_patterns

g = Goose()


def post_processing(sentence: str):
    # handle emojis
    sentence = remove_emoji_patterns(sentence)
    return sentence


def formatting(text: str):
    text = re.sub(r"\n+", "\n", text).strip()
    list_sentences = text.split("\n")
    list_sentences = [post_processing(sentence) for sentence in list_sentences]
    if list_sentences == [""]:
        list_sentences = []
    return list_sentences


def deboilerplating(html: str):
    try:
        result_traf = extract(
            html,
            no_fallback=False,
            include_comments=False,
            include_tables=True,
            include_formatting=False,
        )
        if result_traf is not None:
            result_traf = formatting(result_traf)
        else:
            result_traf = []
    except Exception as e:
        result_traf = []
        logging.info(f"error calling trafilatura: {e}")

    try:
        result_goose = g.extract(raw_html=html)
        result_goose = str(result_goose.cleaned_text)
        if result_goose is not None:
            result_goose = formatting(result_goose)
        else:
            result_goose = []
    except Exception as e:
        result_goose = []
        logging.info(f"error calling goose3: {e}")
    all_sentences = result_traf + result_goose
    sentence_set = list(dict.fromkeys(all_sentences))

    try:
        paragraphs = justext.justext(
            html, justext.get_stoplist("English"), 50, 200, 0.1, 0.2, 0.2, 200, True
        )
        invalid = [
            paragraph.text for paragraph in paragraphs if paragraph.is_boilerplate
        ]
    except Exception as e:
        invalid = []
        logging.info(f"error calling justext: {e}")

    sentence_set = [
        sentence
        for sentence in sentence_set
        if sentence not in invalid or sentence != ""
    ]
    if not sentence_set:
        sentence_set = []
        logging.info("Page not found")

    return dict(
        {
            "set": sentence_set,
            "trafilatura": result_traf,
            "goose": result_goose,
            "justext": invalid,
        }
    )


def get_text_from_html(html: str) -> str:
    result = deboilerplating(html)
    text = " ".join(result["set"])
    return text



# These functions should be reworked in the future to use the companies service

import asyncio
from bson.objectid import ObjectId
from delphai_ml_utils.db import db, db_sync

tech_major_layers = {
    "display": "5fad13fec883a62ab3de75f3",
    "calculated": "5fad13ffc883a62ab3de75f4",
    "added": "5fad13ffc883a62ab3de75f5",
    "removed": "5fad13ffc883a62ab3de75f6",
}
tech_minor_layers = {
    "display": "5fad13ffc883a62ab3de75f7",
    "calculated": "5fad13ffc883a62ab3de75f8",
    "added": "5fad13ffc883a62ab3de75f9",
    "removed": "5fad13ffc883a62ab3de75fa",
}
industry_layers = {
    "display": "5eeb429e0faf396f69e26056",
    "calculated": "5fad13fec883a62ab3de75f0",
    "added": "5fad13fec883a62ab3de75f1",
    "removed": "5fad13fec883a62ab3de75f2",
}
industry_search_layers = {
    "display": "62442ea35e4d371fa97de7ef",
    "calculated": "62442ea35e4d371fa97de7f2",
    "added": "62442ea35e4d371fa97de7f0",
    "removed": "62442ea35e4d371fa97de7f1",
}
layers_mapping = {
    "tech_major": tech_major_layers,
    "tech_minor": tech_minor_layers,
    "industry": industry_layers,
    "industry_search": industry_search_layers,
}


async def get_type_labels(label_type, layer):
    to_return = {}
    async for el in db.labels.find(
        {"layer_id": ObjectId(layers_mapping[label_type][layer])}
    ):
        to_return[el["_id"]] = el["name"]
    return to_return


async def get_company_labels(company_id, labels):
    to_return = []
    async for el in db.companies_labels.find(
        {"company_id": company_id, "label_id": {"$in": list(labels.keys())}},
        sort=[["score", -1], ["rank", -1]],
    ):
        to_return.append(labels[el["label_id"]])
    return to_return


async def get_all_company_labels(comp_id, all_labels):
    company_id = ObjectId(comp_id)
    labels_dict = {"_id": comp_id}
    tasks = []
    for labels in all_labels:
        tasks.append(get_company_labels(company_id, labels))
    awaited = await asyncio.gather(*tasks)
    for layer, response in zip(
        ["industry", "industry_search", "tech_major", "tech_minor"], awaited
    ):
        labels_dict[layer] = response
    return labels_dict


async def get_labels_for_companies(comp_ids, namespace_layer):
    tasks = [
        get_type_labels(label_type, namespace_layer)
        for label_type in ["industry", "industry_search", "tech_major", "tech_minor"]
    ]
    awaited = await asyncio.gather(*tasks)
    industry_labels = awaited[0]
    industry_search_labels = awaited[1]
    tech_major_labels = awaited[2]
    tech_minor_labels = awaited[3]
    all_labels = [
        industry_labels,
        industry_search_labels,
        tech_major_labels,
        tech_minor_labels,
    ]

    tasks = [get_all_company_labels(comp_id, all_labels) for comp_id in comp_ids]
    awaited = await asyncio.gather(*tasks)
    to_return = sorted(awaited, key=lambda x: x["_id"])

    return to_return


def get_labels_for_company(comp_id, namespace_layer):
    company_id = ObjectId(comp_id)
    industry_labels = {
        el["_id"]: el["name"]
        for el in db_sync.labels.find(
            {"layer_id": ObjectId(layers_mapping["industry"][namespace_layer])}
        )
    }
    tech_major_labels = {
        el["_id"]: el["name"]
        for el in db_sync.labels.find(
            {"layer_id": ObjectId(layers_mapping["tech_major"][namespace_layer])}
        )
    }
    tech_minor_labels = {
        el["_id"]: el["name"]
        for el in db_sync.labels.find(
            {"layer_id": ObjectId(layers_mapping["tech_minor"][namespace_layer])}
        )
    }
    industry_search_labels = {
        el["_id"]: el["name"]
        for el in db_sync.labels.find(
            {"layer_id": ObjectId(layers_mapping["industry_search"][namespace_layer])}
        )
    }

    labels_dict = {}
    for layer, labels in (
        ("industry", industry_labels),
        ("tech_major", tech_major_labels),
        ("tech_minor", tech_minor_labels),
        ("industry_search", industry_search_labels),
    ):
        labels_dict[layer] = [
            labels[el["label_id"]]
            for el in db_sync.companies_labels.find(
                {"company_id": company_id, "label_id": {"$in": list(labels.keys())}},
                sort=[["score", -1], ["rank", -1]],
            )
        ]
    return labels_dict


def get_labels_with_ids(label_type, namespace_layer="display"):
    layer_id = ObjectId(layers_mapping[label_type][namespace_layer])
    return {
        el["name"]: str(el["_id"]) for el in db_sync.labels.find({"layer_id": layer_id})
    }



import tldextract
import nltk
import re
from nltk.stem.porter import PorterStemmer
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords

try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("punkt")

try:
    nltk.data.find("corpora/stopwords")
except LookupError:
    nltk.download("stopwords")

stemmer = PorterStemmer()
stop_words = stopwords.words("english")
stop_words.extend(
    [
        "page",
        "app",
        "user",
        "ltd",
        "co",
        "yes",
        "no",
        "pvt",
        "inc",
        "via",
        "refer",
        "websit",
        "site",
        "facebook",
        "twitter",
        "yahoo",
        "instagram",
        "googl",
        "new",
        "com",
        "org",
        "hi",
        "year",
        "ago",
        "wonder",
    ]
)
punctuation_translator = str.maketrans(
    dict.fromkeys("\"#$%&'()*+,-/:;<=>@[\\]^_`{|}~", " ")
)  # .?! removed frpm punctuations
punctuation_mapping = [(" .", "."), (" !", "!"), (" ?", "?")]
remove_parentheses_compile = re.compile(r"""[\(\[].*?[\)\]]""")
remove_trail_space_comma_compile2 = re.compile(r"""\s+,\s+""")
url_compile = r"""(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»""'']))"""
email_compile = r"""(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])"""
phone_compile = r"""(?:(?:\+?([1-9]|[0-9][0-9]|[0-9][0-9][0-9])\s*(?:[.-]\s*)?)?(?:\(\s*([2-9]1[02-9]|[2-9][02-8]1|[2-9][02-8][02-9])\s*\)|([0-9][1-9]|[0-9]1[02-9]|[2-9][02-8]1|[2-9][02-8][02-9]))\s*(?:[.-]\s*)?)?([2-9]1[02-9]|[2-9][02-9]1|[2-9][02-9]{2})\s*(?:[.-]\s*)?([0-9]{4})(?:\s*(?:#|x\.?|ext\.?|extension)\s*(\d+))?"""
contact_email_compile = r"""Contact Email (?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\]) |Official Facebook account for |home \| \- > """
emoji_pattern = re.compile(
    "["
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F1E0-\U0001F1FF"  # flags (iOS)
    "]+",
    flags=re.UNICODE,
)


def remove_emoji_patterns(text: str) -> str:
    return emoji_pattern.sub(r"", text)


def restrict_to_ascii(text: str) -> str:
    return text.encode("ascii", "ignore").decode()


def extract_domain(url: str) -> str:
    # get domain from input e.g. all below would output: nantenergy.com
    # www.nantenergy.com - https://nantenergy.com/ - https://nantenergy.com/ - lauren@nantenegery.com - nantenegery.com
    return tldextract.extract(url).registered_domain


def remove_one_characters(text: str) -> str:
    return re.sub(r"\b[a-zA-Z]\b", "", text)


def remove_numbers(text: str) -> str:
    return re.sub(r" \d+", "", text)


def remove_punctuation(text: str) -> str:
    return text.translate(punctuation_translator)


def remove_stopwords_and_stem(text: str) -> str:
    paras = [p for p in text.split("\n") if p.strip()]
    processed_paras = []
    for para in paras:
        word_tokens = word_tokenize(para)
        # filtered_text = [word for word in word_tokens if word not in stop_words]
        filtered_text = [stemmer.stem(word) for word in word_tokens]
        filtered_text = [word for word in filtered_text if word not in stop_words]
        temp = " ".join(filtered_text)
        processed_paras.append(temp)

    return " \n ".join(processed_paras)


def remove_sub_super_scripts(text: str) -> str:
    return "".join([i for i in text if ord(i) < 128])


def remove_non_ascii(text: str) -> str:
    return "".join(
        char for char in text if ord(char) < 128 and ord(char) not in [8364, 163]
    )


def remove_parantheses(text: str) -> str:
    return re.sub(remove_parentheses_compile, "", text)


def remove_trail_space_comma(text: str) -> str:
    return re.sub(remove_trail_space_comma_compile2, ", ", text)


def replace_sentence_seperators(text: str) -> str:
    for k, v in punctuation_mapping:
        text = text.replace(k, v)
    return text


def detect_urls(text: str) -> list:
    try:
        url = re.findall(url_compile, text)
        urls = [x[0] for x in url]
    except:
        pass
    return urls


def detect_emails(text: str) -> list:
    try:
        email = re.findall(email_compile, text)
        emails = [x[0] for x in email]
    except:
        pass
    return emails


def detect_contact_emails(text: str) -> list:
    try:
        contact_email = re.findall(contact_email_compile, text)
        contact_emails = [x[0] for x in contact_email]
    except:
        pass
    return contact_emails


def detect_phones(text: str) -> list:
    try:
        phone = re.findall(phone_compile, text)
        phones = [x[0] for x in phone]
    except:
        pass
    return phones



import matplotlib.pyplot as plt
import numpy as np
import warnings
from sklearn.metrics import confusion_matrix

# Magnify intervals where font size matters
MAGNIFY_INTERVALS = [0.1, 1]

# Min and max appropriate font sizes
FONT_RANGE = [10.5, 50]

# Maximum allowed magnify. This will get multiplied by 0 - 1 value.
MAX_MAGNIFY = 15

# Increase font for title ratio.
TITLE_FONT_RATIO = 1.8


def plot_array(
    array,
    start_step=0,
    step_size=1,
    use_label=None,
    use_title=None,
    points_values=False,
    points_round=3,
    use_xlabel=None,
    use_xticks=True,
    use_ylabel=None,
    style_sheet="ggplot",
    use_grid=True,
    use_linestyle="-",
    font_size=None,
    width=3,
    height=1,
    magnify=0.1,
    use_dpi=50,
    path=None,
    show_plot=True,
):
    r"""
    Create plot from a single array of values.
    Arguments:
        array (:obj:`list / np.ndarray`):
            List of values. Can be of type list or np.ndarray.
        start_step (:obj:`int`, `optional`, defaults to :obj:`0`):
            Value to offset all x-axis values. This argument is optional and it has a default value attributed inside
            the function.
        step_size (:obj:`int`, `optional`, defaults to :obj:`1`):
            What is the step increase of each point on the x axis.
            This argument is optional and it has a default value attributed inside the function. It will multiply
            each x-axis position value to it.
        use_label (:obj:`str`, `optional`):
            Display label on plot for whole array. This argument is optional and it will have a `None` value attributed
            inside the function.
        use_title (:obj:`str`, `optional`):
            Title on top of plot. This argument is optional and it will have a `None` value attributed inside the
            function for which it won't display any title.
        points_values (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Display each point value on the plot. This argument is optional and it has a default value attributed
            inside the function.
        points_round (:obj:`int`, `optional`, defaults to :obj:`1`):
            Round decimal values for points values. This argument is optional and it has a default value attributed
            inside the function.
        use_xlabel (:obj:`str`, `optional`):
            Label to use for x-axis value meaning. This argument is optional and it will have a `None` value attributed
            inside the function.
        use_xticks (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Display x-axis tick values (the values at each point). This argument is optional and it has a default
            value attributed inside the function.
        use_ylabel (:obj:`str`, `optional`):
            Label to use for y-axis value meaning. This argument is optional and it will have a `None` value attributed
            inside the function.
        style_sheet (:obj:`str`, `optional`, defaults to :obj:`ggplot`):
            Style of plot. Use plt.style.available to show all styles. This argument is optional and it has a default
            value attributed inside the function.
        use_grid (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Show grid on plot or not. This argument is optional and it has a default value attributed inside
            the function.
        use_linestyle (:obj:`str`, `optional`, defaults to :obj:`-`):
            Style to use on line from ['-', '--', '-.', ':']. This argument is optional and it has a default
            value attributed inside the function.
        font_size (:obj:`int` or `float`, `optional`):
            Font size to use across the plot. By default this function will adjust font size depending on `magnify`
            value. If this value is set, it will ignore the `magnify` recommended font size. The title font size is by
            default `1.8` greater than font-size. This argument is optional and it will have a `None` value attributed
            inside the function.
        width (:obj:`int`, `optional`, defaults to :obj:`3`):
            Horizontal length of plot. This argument is optional and it has a default value attributed inside
            the function.
        height (:obj:`int`, `optional`, defaults to :obj:`1`):
            Height length of plot in inches. This argument is optional and it has a default value attributed inside
            the function.
        magnify (:obj:`float`, `optional`, defaults to :obj:`0.1`):
            Ratio increase of both with and height keeping the same ratio size. This argument is optional and it has a
            default value attributed inside the function.
        use_dpi (:obj:`int`, `optional`, defaults to :obj:`50`):
            Print resolution is measured in dots per inch (or “DPI”). This argument is optional and it has a default
            value attributed inside the function.
        path (:obj:`str`, `optional`):
            Path and file name of plot saved as image. If want to save in current path just pass in the file name.
            This argument is optional and it will have a None value attributed inside the function.
        show_plot (:obj:`bool`, `optional`, defaults to :obj:`1`):
            if you want to call `plt.show()`. or not (if you run on a headless server). This argument is optional and
            it has a default value attributed inside the function.
    Raises:
        ValueError: If `array` is not of type `list` or `np.ndarray`
        ValueError: If `style_sheet` is not valid.
        ValueError: If `use_linestyle` is not valid.
        DeprecationWarning: If `magnify` is se to values that don't belong to [0, 1] values.
        ValueError: If `font_size` is not `None` and smaller or equal to 0.
    """

    # Check if `array` is in correct format.
    if not isinstance(array, list) or isinstance(array, np.ndarray):
        # raise value error.
        raise ValueError("`array` needs to be a list of values!")

    # Check if `style_sheet` has correct value.
    if style_sheet in plt.style.available:
        # Set style of plot
        plt.style.use(style_sheet)
    else:
        # Style is not correct.
        raise ValueError(
            "`style_sheet=%s` is not in the supported styles: %s"
            % (str(style_sheet), str(plt.style.available))
        )
    # Make sure `magnify` is in right range.
    if magnify > 1 or magnify <= 0:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify` needs to have value in [0,1]! `{magnify}` will be converted to `0.1` as default.",
            DeprecationWarning,
        )
        # Convert to regular value 0.1.
        magnify = 0.1

    # All allowed linestyles.
    linestyles = ["-", "--", "-.", ":"]

    # Make sure `font_size` is set right.
    if (font_size is not None) and (font_size <= 0):
        # Raise value error -  is not correct.
        raise ValueError(
            f"`font_size` needs to be positive number! Invalid value {font_size}"
        )

    # Font size select custom or adjusted on `magnify` value.
    font_size = (
        font_size
        if font_size is not None
        else np.interp(magnify, MAGNIFY_INTERVALS, FONT_RANGE)
    )

    # Font variables dictionary. Keep it in this format for future updates.
    font_dict = dict(
        family="DejaVu Sans",
        color="black",
        weight="normal",
        size=font_size,
    )

    # Check if linestyle is set right.
    if use_linestyle not in linestyles:
        # Raise error.
        raise ValueError(
            "`linestyle=%s` is not in the styles: %s!"
            % (str(use_linestyle), str(linestyles))
        )

    # Set steps plotted on x-axis - we can use step if 1 unit has different value.
    if start_step > 0:
        # Offset all steps by start_step.
        steps = np.array(range(0, len(array))) * step_size + start_step
    else:
        # Keep steps the same.
        steps = np.array(range(1, len(array) + 1)) * step_size

    # Single plot figure.
    plt.subplot(1, 2, 1)

    # Plot array as a single line
    plt.plot(steps, array, linestyle=use_linestyle, label=use_label)

    # Plots points values.
    if points_values:
        # Loop through each point and plot the label.
        for x, y in zip(steps, array):
            # Add text label to plot.
            plt.text(x, y, str(round(y, points_round)), fontdict=font_dict)

    # Set horizontal axis name.
    plt.xlabel(use_xlabel, fontdict=font_dict)

    # Use x ticks with steps.
    plt.xticks(steps) if use_xticks else None

    # Set vertical axis name.
    plt.ylabel(use_ylabel, fontdict=font_dict)

    # Place legend best position.
    plt.legend(
        loc="best", fontsize=font_dict["size"]
    ) if use_label is not None else None

    # Custom label font size for x axis. This is for future updates if needed.
    # plt.tick_params(axis="x", labelsize=font['size']//2)

    # Custom label font size for y axis. This is for future updates if needed.
    # plt.tick_params(axis="y", labelsize=font['size']//2)

    # Adjust both axis labels font size at same time.
    plt.tick_params(labelsize=font_dict["size"])

    # Display grid depending on `use_grid`.
    plt.grid(use_grid)

    # Make figure nice.
    plt.tight_layout()

    # Adjust font for
    font_dict["size"] *= TITLE_FONT_RATIO

    # Set title of figure.
    plt.title(use_title, fontdict=font_dict)

    # Rescale `magnify` to be used on inches.
    magnify *= MAX_MAGNIFY

    # Get figure object from plot.
    fig = plt.gcf()

    # Get size of figure.
    figsize = fig.get_size_inches()

    # Change size depending on height and width variables.
    figsize = [figsize[0] * width * magnify, figsize[1] * height * magnify]

    # Set the new figure size.
    fig.set_size_inches(figsize)

    # There is an error when DPI and plot size are too large!
    try:
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None
    except ValueError:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify={magnify // 15}` is to big in combination"
            f" with `use_dpi={use_dpi}`! Try using lower values for"
            f" `magnify` and/or `use_dpi`. Image was saved in {path}"
            f" with `use_dpi=50 and `magnify={magnify // 15}`!",
            Warning,
        )
        # Set DPI to smaller value and warn user to use smaller magnify or smaller dpi.
        use_dpi = 50
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None

    # Show plot.
    plt.show() if show_plot is True else None

    return


def plot_dict(
    dict_arrays,
    start_step=0,
    step_size=1,
    use_title=None,
    points_values=False,
    points_round=3,
    use_xlabel=None,
    use_xticks=True,
    use_rotation_xticks=0,
    xticks_labels=None,
    use_ylabel=None,
    style_sheet="ggplot",
    use_grid=True,
    use_linestyles=None,
    font_size=None,
    width=3,
    height=1,
    magnify=1.2,
    use_dpi=50,
    path=None,
    show_plot=True,
):
    r"""
    Create plot from a single array of values.
    Arguments:
        dict_arrays (:obj:`dict([list])`):
            Dictionary of arrays that will get plotted. The keys in dictionary are used as labels and the values as
            arrays that get plotted.
        start_step (:obj:`int`, `optional`, defaults to :obj:`0`):
            Starting value of plot.This argument is optional and it has a default value attributed inside
            the function.
        step_size (:obj:`int`, `optional`, defaults to :obj:`q`):
            Steps shows on x-axis. Change if each steps is different than 1.This argument is optional and it has a
            default value attributed inside the function.
        use_title (:obj:`int`, `optional`):
            Title on top of plot. This argument is optional and it will have a `None` value attributed
            inside the function.
        points_values (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Display each point value on the plot. This argument is optional and it has a default value attributed
            inside the function.
        points_round (:obj:`int`, `optional`, defaults to :obj:`1`):
            Round decimal valus for points values. This argument is optional and it has a default value attributed
            inside the function.
        use_xlabel (:obj:`str`, `optional`):
            Label to use for x-axis value meaning. This argument is optional and it will have a `None` value attributed
            inside the function.
        use_xticks (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Display x-axis tick values (the values at each point). This argument is optional and it has a default
            value attributed inside the function.
        use_ylabel (:obj:`str`, `optional`):
            Label to use for y-axis value meaning. This argument is optional and it will have a `None` value attributed
            inside the function.
        style_sheet (:obj:`str`, `optional`, defaults to :obj:`ggplot`):
            Style of plot. Use plt.style.available to show all styles. This argument is optional and it has a default
            value attributed inside the function.
        use_grid (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Show grid on plot or not. This argument is optional and it has a default value attributed inside
            the function.
        use_linestyles (:obj:`str`, `optional`, defaults to :obj:`-`):
            Style to use on line from ['-', '--', '-.', ':']. This argument is optional and it has a default
            value attributed inside the function.
        font_size (:obj:`int` or `float`, `optional`):
            Font size to use across the plot. By default this function will adjust font size depending on `magnify`
            value. If this value is set, it will ignore the `magnify` recommended font size. The title font size is by
            default `1.8` greater than font-size. This argument is optional and it will have a `None` value attributed
            inside the function.
        width (:obj:`int`, `optional`, defaults to :obj:`3`):
            Horizontal length of plot. This argument is optional and it has a default value attributed inside
            the function.
        height (:obj:`int`, `optional`, defaults to :obj:`1`):
            Height length of plot in inches. This argument is optional and it has a default value attributed inside
            the function.
        magnify (:obj:`float`, `optional`, defaults to :obj:`0.1`):
            Ratio increase of both with and height keeping the same ratio size. This argument is optional and it has a
            default value attributed inside the function.
        use_dpi (:obj:`int`, `optional`, defaults to :obj:`50`):
            Print resolution is measured in dots per inch (or “DPI”). This argument is optional and it has a default
            value attributed inside the function.
        path (:obj:`str`, `optional`):
            Path and file name of plot saved as image. If want to save in current path just pass in the file name.
            This argument is optional and it will have a None value attributed inside the function.
        show_plot (:obj:`bool`, `optional`, defaults to :obj:`1`):
            if you want to call `plt.show()`. or not (if you run on a headless server). This argument is optional and
            it has a default value attributed inside the function.
    Raises:
        ValueError: If `dict_arrays` is not of type `dictionary`.
        ValueError: If `dict_arrays` doesn't have string keys.
        ValueError: If `dict_arrays` doesn't have array values.
        ValueError: If `style_sheet` is not valid.
        ValueError: If `use_linestyle` is not valid.
        ValueError: If `points_values`of type list don't have same length as `dict_arrays`.
        DeprecationWarning: If `magnify` is se to values that don't belong to [0, 1] values.
        ValueError: If `font_size` is not `None` and smaller or equal to 0.
    """

    # Check if `dict_arrays` is the correct format.
    if not isinstance(dict_arrays, dict):
        # Raise value error.
        raise ValueError("`dict_arrays` needs to be a dictionary of values!")

    # Check each label
    for label, array in dict_arrays.items():
        # Check if format is correct.
        if not isinstance(label, str):
            # Raise value error.
            raise ValueError("`dict_arrays` needs string keys!")
        if not isinstance(array, list) or isinstance(array, np.ndarray):
            # Raise value error.
            raise ValueError("`dict_arrays` needs lists values!")

    # Make sure style sheet is correct.
    if style_sheet in plt.style.available:
        # Set style of plot
        plt.style.use(style_sheet)
    else:
        # Style is not correct.
        raise ValueError(
            "`style_sheet=%s` is not in the supported styles: %s"
            % (str(style_sheet), str(plt.style.available))
        )

    # Make sure `magnify` is in right range.
    if magnify > 1 or magnify <= 0:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify` needs to have value in [0,1]! `{magnify}` will be converted to `0.1` as default.",
            DeprecationWarning,
        )
        # Convert to regular value 0.1.
        magnify = 0.1

    # all linestyles.
    linestyles = ["-", "--", "-.", ":"]

    # Make sure `font_size` is set right.
    if (font_size is not None) and (font_size <= 0):
        # Raise value error -  is not correct.
        raise ValueError(
            f"`font_size` needs to be positive number! Invalid value {font_size}"
        )

    # Font size select custom or adjusted on `magnify` value.
    font_size = (
        font_size
        if font_size is not None
        else np.interp(magnify, MAGNIFY_INTERVALS, FONT_RANGE)
    )

    # Font variables dictionary. Keep it in this format for future updates.
    font_dict = dict(
        family="DejaVu Sans",
        color="black",
        weight="normal",
        size=font_size,
    )

    # If single style value is passed, use it on all arrays.
    if use_linestyles is None:
        use_linestyles = ["-"] * len(dict_arrays)

    else:
        # Check if linestyle is set right.
        for use_linestyle in use_linestyles:
            if use_linestyle not in linestyles:
                # Raise error.
                raise ValueError(
                    "`linestyle=%s` is not in the styles: %s!"
                    % (str(use_linestyle), str(linestyles))
                )

    # Check `points_value` type - it can be bool or list(bool).
    if isinstance(points_values, bool):
        # Convert to list.
        points_values = [points_values] * len(dict_arrays)
    elif isinstance(points_values, list) and (len(points_values) != len(dict_arrays)):
        raise ValueError(
            "`points_values` of type `list` must have same length as dictionary!"
        )

    # Single plot figure.
    plt.subplot(1, 2, 1)

    # Use maximum length of steps. In case each arrya has different lengths.
    max_steps = []

    # Plot each array.
    for index, (use_label, array) in enumerate(dict_arrays.items()):
        # Set steps plotted on x-axis - we can use step if 1 unit has different value.
        if start_step > 0:
            # Offset all steps by start_step.
            steps = np.array(range(0, len(array))) * step_size + start_step
            max_steps = steps if len(max_steps) < len(steps) else max_steps
        else:
            steps = np.array(range(1, len(array) + 1)) * step_size
            max_steps = steps if len(max_steps) < len(steps) else max_steps

        # Plot array as a single line.
        plt.plot(steps, array, linestyle=use_linestyles[index], label=use_label)

        # Plots points values.
        if points_values[index]:
            # Loop through each point and plot the label.
            for x, y in zip(steps, array):
                # Add text label to plot.
                plt.text(x, y, str(round(y, points_round)), fontdict=font_dict)

    # Set horizontal axis name.
    plt.xlabel(use_xlabel, fontdict=font_dict)

    # Use x ticks with steps or labels.
    plt.xticks(
        max_steps, xticks_labels, rotation=use_rotation_xticks
    ) if use_xticks else None

    # Set vertical axis name.
    plt.ylabel(use_ylabel, fontdict=font_dict)

    # Adjust both axis labels font size at same time.
    plt.tick_params(labelsize=font_dict["size"])

    # Place legend best position.
    plt.legend(loc="best", fontsize=font_dict["size"])

    # Adjust font for title.
    font_dict["size"] *= TITLE_FONT_RATIO

    # Set title of figure.
    plt.title(use_title, fontdict=font_dict)

    # Rescale `magnify` to be used on inches.
    magnify *= MAX_MAGNIFY

    # Display grid depending on `use_grid`.
    plt.grid(use_grid)

    # Make figure nice.
    plt.tight_layout()

    # Get figure object from plot.
    fig = plt.gcf()

    # Get size of figure.
    figsize = fig.get_size_inches()

    # Change size depending on height and width variables.
    figsize = [figsize[0] * width * magnify, figsize[1] * height * magnify]

    # Set the new figure size with magnify.
    fig.set_size_inches(figsize)

    # There is an error when DPI and plot size are too large!
    try:
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None
    except ValueError:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify={magnify // 15}` is to big in combination"
            f" with `use_dpi={use_dpi}`! Try using lower values for"
            f" `magnify` and/or `use_dpi`. Image was saved in {path}"
            f" with `use_dpi=50 and `magnify={magnify // 15}`!",
            Warning,
        )
        # Set DPI to smaller value and warn user to use smaller magnify or smaller dpi.
        use_dpi = 50
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None

    # Show plot.
    plt.show() if show_plot is True else None

    return


def plot_confusion_matrix(
    y_true,
    y_pred,
    use_title=None,
    classes="",
    normalize=False,
    style_sheet="ggplot",
    cmap=plt.cm.Blues,
    font_size=None,
    verbose=0,
    width=3,
    height=1,
    magnify=0.1,
    use_dpi=50,
    path=None,
    show_plot=True,
    **kwargs,
):
    r"""
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    Arguments:
        y_true (:obj:`list / np.ndarray`):
            List of labels values.
        y_pred (:obj:`list / np.ndarray`):
            List of predicted label values.
        use_title (:obj:`int`, `optional`):
            Title on top of plot. This argument is optional and it will have a `None` value attributed
            inside the function.
        classes (:obj:`str`, `optional`, defaults to :obj:``):
            List of label names. This argument is optional and it has a default value attributed
            inside the function.
        normalize (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Normalize confusion matrix or not. This argument is optional and it has a default value attributed
            inside the function.
        style_sheet (:obj:`str`, `optional`, defaults to :obj:`ggplot`):
            Style of plot. Use plt.style.available to show all styles. This argument is optional and it has a default
            value attributed inside the function.
        cmap (:obj:`str`, `optional`, defaults to :obj:`plt.cm.Blues`):
            It is a plt.cm plot theme. Plot themes: `plt.cm.Blues`, `plt.cm.BuPu`, `plt.cm.GnBu`, `plt.cm.Greens`,
            `plt.cm.OrRd`. This argument is optional and it has a default value attributed inside the function.
        font_size (:obj:`int` or `float`, `optional`):
            Font size to use across the plot. By default this function will adjust font size depending on `magnify`
            value. If this value is set, it will ignore the `magnify` recommended font size. The title font size is by
            default `1.8` greater than font-size. This argument is optional and it will have a `None` value attributed
            inside the function.
        verbose (:obj:`int`, `optional`, defaults to :obj:`0`):
            To display confusion matrix value or not if set > 0. This argument is optional and it has a default
            value attributed inside the function.
        width (:obj:`int`, `optional`, defaults to :obj:`3`):
            Horizontal length of plot. This argument is optional and it has a default value attributed inside
            the function.
        height (:obj:`int`, `optional`, defaults to :obj:`1`):
            Height length of plot in inches. This argument is optional and it has a default value attributed inside
            the function.
        magnify (:obj:`float`, `optional`, defaults to :obj:`0.1`):
            Ratio increase of both with and height keeping the same ratio size. This argument is optional and it has a
            default value attributed inside the function.
        use_dpi (:obj:`int`, `optional`, defaults to :obj:`50`):
            Print resolution is measured in dots per inch (or “DPI”). This argument is optional and it has a default
            value attributed inside the function.
        path (:obj:`str`, `optional`):
            Path and file name of plot saved as image. If want to save in current path just pass in the file name.
            This argument is optional and it will have a None value attributed inside the function.
        show_plot (:obj:`bool`, `optional`, defaults to :obj:`1`):
            if you want to call `plt.show()`. or not (if you run on a headless server). This argument is optional and
            it has a default value attributed inside the function.
        kwargs (:obj:`dict`, `optional`):
            Other arguments that might be deprecated or not included as details. This argument is optional and it will
            have a `None` value attributed inside the function.
    Returns:
        :obj:`np.ndarray`: Confusion matrix used to plot.
    Raises:
        DeprecationWarning: If arguments `title` is used.
        DeprecationWarning: If arguments `image` is used.
        DeprecationWarning: If arguments `dpi` is used.
        ValueError: If `y_true` and `y_pred` arrays don't have same length.
        ValueError: If `dict_arrays` doesn't have string keys.
        ValueError: If `dict_arrays` doesn't have array values.
        ValueError: If `style_sheet` is not valid.
        DeprecationWarning: If `magnify` is se to values that don't belong to [0, 1] values.
        ValueError: If `font_size` is not `None` and smaller or equal to 0.
    """

    # Handle deprecation warnings if `title` is used.
    if "title" in kwargs:
        # assign same value
        use_title = kwargs["title"]
        warnings.warn(
            "`title` will be deprecated in future updates. Use `use_title` in stead!",
            DeprecationWarning,
        )

    # Handle deprecation warnings if `image` is used.
    if "image" in kwargs:
        # assign same value
        path = kwargs["image"]
        warnings.warn(
            "`image` will be deprecated in future updates. Use `path` in stead!",
            DeprecationWarning,
        )
    # Handle deprecation warnings if `dpi` is used.
    if "dpi" in kwargs:
        # assign same value
        use_dpi = kwargs["dpi"]
        warnings.warn(
            "`dpi` will be deprecated in future updates. Use `use_dpi` in stead!",
            DeprecationWarning,
        )

    # Make sure labels have right format.
    if len(y_true) != len(y_pred):
        # make sure lengths match
        raise ValueError("`y_true` needs to have same length as `y_pred`!")

    # Make sure style sheet is correct.
    if style_sheet in plt.style.available:
        # set style of plot
        plt.style.use(style_sheet)
    else:
        # style is not correct
        raise ValueError(
            "`style_sheet=%s` is not in the supported styles: %s"
            % (str(style_sheet), str(plt.style.available))
        )

    # Make sure `magnify` is in right range.
    if magnify > 1 or magnify <= 0:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify` needs to have value in [0,1]! `{magnify}` will be converted to `0.1` as default.",
            DeprecationWarning,
        )
        # Convert to regular value 0.1.
        magnify = 0.1

    # Make sure `font_size` is set right.
    if (font_size is not None) and (font_size <= 0):
        # Raise value error -  is not correct.
        raise ValueError(
            f"`font_size` needs to be positive number! Invalid value {font_size}"
        )

    # Font size select custom or adjusted on `magnify` value.
    font_size = (
        font_size
        if font_size is not None
        else np.interp(magnify, MAGNIFY_INTERVALS, FONT_RANGE)
    )

    # Font variables dictionary. Keep it in this format for future updates.
    font_dict = dict(
        family="DejaVu Sans",
        color="black",
        weight="normal",
        size=font_size,
    )

    # Class labels setup. If none, generate from y_true y_pred.
    classes = list(classes)
    if classes:
        assert len(set(y_true)) == len(classes)
    else:
        classes = set(y_true)

    # Compute confusion matrix.
    cm = confusion_matrix(y_true, y_pred)

    # Normalize setup.
    if normalize is True:
        print("Normalized confusion matrix")
        cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]
        use_title = "Normalized confusion matrix" if use_title is None else use_title
    else:
        print("Confusion matrix, without normalization")
        use_title = (
            "Confusion matrix, without normalization"
            if use_title is None
            else use_title
        )

    # Print if verbose.
    print(cm) if verbose > 0 else None

    # Plot setup.
    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
    ax.figure.colorbar(im, ax=ax)

    # Show all ticks.
    ax.set(
        xticks=np.arange(cm.shape[1]),
        yticks=np.arange(cm.shape[0]),
        # Label ticks with the respective list entries.
        xticklabels=classes,
        yticklabels=classes,
    )

    # Set horizontal axis name.
    ax.set_xlabel("Predicted label", fontdict=font_dict)

    # Set vertical axis name.
    ax.set_ylabel("True label", fontdict=font_dict)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
    plt.grid(False)

    # Loop over data dimensions and create text annotations.
    fmt = ".2f" if normalize else "d"
    thresh = cm.max() / 2.0
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(
                j,
                i,
                format(cm[i, j], fmt),
                ha="center",
                va="center",
                color="white" if cm[i, j] > thresh else "black",
                fontdict=font_dict,
            )

    # Adjust both axis labels font size at same time.
    plt.tick_params(labelsize=font_dict["size"])

    # Adjust font for title.
    font_dict["size"] *= TITLE_FONT_RATIO

    # Set title of figure.
    plt.title(use_title, fontdict=font_dict)

    # Rescale `magnify` to be used on inches.
    magnify *= MAX_MAGNIFY

    # Never display grid.
    plt.grid(False)

    # Make figure nice.
    plt.tight_layout()

    # Get figure object from plot.
    fig = plt.gcf()

    # Get size of figure.
    figsize = fig.get_size_inches()

    # Change size depending on height and width variables.
    figsize = [figsize[0] * width * magnify, figsize[1] * height * magnify]

    # Set the new figure size with magnify.
    fig.set_size_inches(figsize)

    # There is an error when DPI and plot size are too large!
    try:
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None
    except ValueError:
        # Deprecation warning from last time.
        warnings.warn(
            f"`magnify={magnify // 15}` is to big in combination"
            f" with `use_dpi={use_dpi}`! Try using lower values for"
            f" `magnify` and/or `use_dpi`. Image was saved in {path}"
            f" with `use_dpi=50 and `magnify={magnify // 15}`!",
            Warning,
        )
        # Set DPI to smaller value and warn user to use smaller magnify or smaller dpi.
        use_dpi = 50
        # Save figure to image if path is set.
        fig.savefig(
            path, dpi=use_dpi, bbox_inches="tight"
        ) if path is not None else None

    # Show plot.
    plt.show() if show_plot is True else None

    return cm



import nlpaug.augmenter.word as naw
import pysbd
import random
import numpy as np
import nltk
import torch
import logging

try:
    nltk.data.find("taggers/averaged_perceptron_tagger")
except LookupError:
    nltk.download("averaged_perceptron_tagger")

try:
    nltk.data.find("corpora/omw-1.4")
except LookupError:
    nltk.download("omw-1.4")

try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("wordnet")


class DataAugmentation:

    """
    augment text by companies names random shuffling, back-translation and words synonyms
    """

    def __init__(
        self, translation_model_config_by_name=dict(), all_ner_mentions_pool=set()
    ):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        logging.info(f"Device: {self.device}")

        self.ner_mentions_pool = all_ner_mentions_pool  # can either input list of all ner mentions or just add iteratively

        self.translation_model_by_name = dict()
        if not translation_model_config_by_name:
            logging.warning(
                """No back-translations have been given as input. If you are planning to use
        the back-translation augmentation, you need to add this in when initialising the module. 
        Example: translation_model_config_by_name=
        {
          'de': 
            {
                'from_model': 'Helsinki-NLP/opus-mt-en-de',
                'to_model': 'Helsinki-NLP/opus-mt-de-en'
            }
        }

        Find available models here: https://huggingface.co/Helsinki-NLP"""
            )
        else:
            self._download_back_translation_models(translation_model_config_by_name)

        synonym_regex = "NE[0-9]*"  # pattern for indicating companies names
        self.synonym_augmenter = naw.SynonymAug(
            aug_src="wordnet", aug_p=0.3, stopwords_regex=synonym_regex
        )  # initiate augmenter with synonyms (using companies names pattern as stopwords)

        # (translation on the whole text is not stable, sometimes it translates only two-three sentences)
        self.seg = pysbd.Segmenter(language="en", char_span=True)

    def _create_mask_mapping(self, ner_mentions: list) -> dict:
        """to build mask mapping dictionary for a row"""
        mask_mapping = dict()
        for i, name in enumerate(ner_mentions):
            self.ner_mentions_pool.add(name)
            mask_mapping[name] = f"NE{str(i)}"
        return mask_mapping

    def _mask_company_names(self, text: str, mask_mapping: dict) -> tuple:
        """# for masking companies names"""
        for entity_name, mask in list(mask_mapping.items()):
            text = text.replace(entity_name, mask)
        return text

    def _unmask_company_names(self, text: str, mask_mapping: dict) -> str:
        """for unmasking companies names"""
        for entity_name, mask in list(mask_mapping.items()):
            text = text.replace(mask, entity_name)
        return text

    def _download_back_translation_models(
        self, translation_model_config_by_name: dict
    ) -> None:
        logging.info("Donwloading back-translation models")
        for k, v in translation_model_config_by_name.items():
            logging.info(f"""From: {v["from_model"]}, To: {v["to_model"]}""")
            try:
                back_translation_aug_model = naw.BackTranslationAug(
                    from_model_name=v["from_model"],
                    to_model_name=v["to_model"],
                    device=self.device,
                )
                self.translation_model_by_name[k] = back_translation_aug_model
            except Exception as e:
                logging.error(e)

    def augmentation_by_back_translation(
        self, text: str, mask_mapping: dict, sentence_segmentation=True
    ) -> tuple:
        """
        Output augmented text result and language used (None for both if requirements
        not met.)
        List of back-translations corresponds to languages with which translations were done
        """
        if not self.translation_model_by_name:
            logging.error(
                f"""There are no downloaded back-translation models. Please re-initialise the module"""
            )

        unmasked_texts = []
        aug_langs = []
        masked_text = self._mask_company_names(text, mask_mapping)

        for lang, back_translation_aug in self.translation_model_by_name.items():
            if sentence_segmentation:
                augmented_text = " ".join(
                    [
                        back_translation_aug.augment(sent)[0]
                        for sent in self.seg.processor(masked_text).process()
                    ]
                )
            else:
                augmented_text = back_translation_aug.augment(masked_text)[0]
                # coin flip for random company names replacement
            if random.random() > 0.5:
                new_names = np.random.choice(
                    list(self.ner_mentions_pool), size=len(mask_mapping), replace=False
                )
                mask_mapping = dict(zip(new_names, mask_mapping.values()))
            unmasked_text = self._unmask_company_names(augmented_text, mask_mapping)
            if unmasked_text != text:
                unmasked_texts.append(unmasked_text)
                aug_langs.append(lang)

        return unmasked_texts, aug_langs

    def augmentation_by_shuffling(self, text: str, mask_mapping: dict) -> tuple:
        """
        Entities in original text are randomly replaced by others from the pool.
        """
        masked_text = self._mask_company_names(text, mask_mapping)
        new_names = np.random.choice(
            list(self.ner_mentions_pool), size=len(mask_mapping), replace=False
        )
        mask_mapping = dict(zip(new_names, mask_mapping.values()))
        augmented_text = self._unmask_company_names(masked_text, mask_mapping)
        return augmented_text

    def augmentation_by_synonyms(self, text: str, mask_mapping: dict) -> tuple:
        """
        Random words are replaced by synonyms
        """
        masked_text = self._mask_company_names(text, mask_mapping)
        augmented_text_result = None
        augmented_text = self.synonym_augmenter.augment(masked_text)[0]
        augmented_text = (
            augmented_text.replace(" ’ ", "’")
            .replace("“ ", "“")
            .replace(" ”", "”")
            .replace(" ' ", "'")
        )

        # coin flip for random company names replacement
        if random.random() > 0.5:
            new_names = np.random.choice(
                list(self.ner_mentions_pool), size=len(mask_mapping), replace=False
            )
            mask_mapping = dict(zip(new_names, mask_mapping.values()))

        unmasked_text = self._unmask_company_names(augmented_text, mask_mapping)
        if unmasked_text != text:
            augmented_text_result = unmasked_text
        return augmented_text_result  # will be None if requirements are not met



import torch
import torch.nn as nn


"""
Asymmetric Loss for multi-label classification
code source - https://github.com/Alibaba-MIIL/ASL/blob/main/src/loss_functions/losses.py

paper - https://paperswithcode.com/paper/asymmetric-loss-for-multi-label
"""


class AsymmetricLoss(nn.Module):
    def __init__(
        self,
        gamma_neg=4,
        gamma_pos=1,
        clip=0.05,
        eps=1e-8,
        disable_torch_grad_focal_loss=True,
    ):
        super(AsymmetricLoss, self).__init__()

        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.clip = clip
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.eps = eps

    def forward(self, x, y):
        """ "
        Parameters
        ----------
        x: input logits
        y: targets (multi-label binarized vector)
        """

        # Calculating Probabilities
        x_sigmoid = torch.sigmoid(x)
        xs_pos = x_sigmoid
        xs_neg = 1 - x_sigmoid

        # Asymmetric Clipping
        if self.clip is not None and self.clip > 0:
            xs_neg = (xs_neg + self.clip).clamp(max=1)

        # Basic CE calculation
        los_pos = y * torch.log(xs_pos.clamp(min=self.eps))
        los_neg = (1 - y) * torch.log(xs_neg.clamp(min=self.eps))
        loss = los_pos + los_neg

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(False)
            pt0 = xs_pos * y
            pt1 = xs_neg * (1 - y)  # pt = p if t > 0 else 1-p
            pt = pt0 + pt1
            one_sided_gamma = self.gamma_pos * y + self.gamma_neg * (1 - y)
            one_sided_w = torch.pow(1 - pt, one_sided_gamma)
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(True)
            loss *= one_sided_w

        return -loss.sum()


class AsymmetricLossOptimized(nn.Module):
    """Notice - optimized version, minimizes memory allocation and gpu uploading,
    favors inplace operations"""

    def __init__(
        self,
        gamma_neg=4,
        gamma_pos=1,
        clip=0.05,
        eps=1e-8,
        disable_torch_grad_focal_loss=False,
    ):
        super(AsymmetricLossOptimized, self).__init__()

        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.clip = clip
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.eps = eps

        # prevent memory allocation and gpu uploading every iteration, and encourages inplace operations
        self.targets = (
            self.anti_targets
        ) = self.xs_pos = self.xs_neg = self.asymmetric_w = self.loss = None

    def forward(self, x, y):
        """ "
        Parameters
        ----------
        x: input logits
        y: targets (multi-label binarized vector)
        """

        self.targets = y
        self.anti_targets = 1 - y

        # Calculating Probabilities
        self.xs_pos = torch.sigmoid(x)
        self.xs_neg = 1.0 - self.xs_pos

        # Asymmetric Clipping
        if self.clip is not None and self.clip > 0:
            self.xs_neg.add_(self.clip).clamp_(max=1)

        # Basic CE calculation
        self.loss = self.targets * torch.log(self.xs_pos.clamp(min=self.eps))
        self.loss.add_(self.anti_targets * torch.log(self.xs_neg.clamp(min=self.eps)))

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(False)
            self.xs_pos = self.xs_pos * self.targets
            self.xs_neg = self.xs_neg * self.anti_targets
            self.asymmetric_w = torch.pow(
                1 - self.xs_pos - self.xs_neg,
                self.gamma_pos * self.targets + self.gamma_neg * self.anti_targets,
            )
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(True)
            self.loss *= self.asymmetric_w

        return -self.loss.sum()



import logging
import os
from base64 import b64decode
from time import sleep

from dotenv import load_dotenv
from omegaconf import OmegaConf
from memoization import cached

logger = logging.getLogger(__name__)


env_path = os.path.abspath(".env")
load_dotenv(dotenv_path=env_path)

CURRENT_CONFIG = None
CONFIG_IS_LOADING = False


def __load_config(config_dir: str = "./config"):
    global CURRENT_CONFIG, CONFIG_IS_LOADING
    while CONFIG_IS_LOADING:
        sleep(0.05)
    if CURRENT_CONFIG is not None:
        return CURRENT_CONFIG
    while CONFIG_IS_LOADING:
        sleep(0.05)
    CONFIG_IS_LOADING = True
    try:
        if "DELPHAI_ENVIRONMENT" not in os.environ:
            raise Exception("DELPHAI_ENVIRONMENT is not defined")
        default_config = OmegaConf.load(f"{config_dir}/default.yml")
        delphai_environment = os.environ.get("DELPHAI_ENVIRONMENT")
        if os.path.isfile(f"{config_dir}/{delphai_environment}.yml"):
            delphai_env_config = OmegaConf.load(
                f"{config_dir}/{delphai_environment}.yml"
            )
        else:
            delphai_env_config = OmegaConf.create()
        CURRENT_CONFIG = OmegaConf.merge(default_config, delphai_env_config)
        OmegaConf.set_readonly(CURRENT_CONFIG, True)
        return CURRENT_CONFIG
    finally:
        CONFIG_IS_LOADING = False


@cached
def get_config(path: str = "", config_dir: str = "./config"):
    config = __load_config(config_dir=config_dir)
    if path is None:
        return config
    selected = OmegaConf.select(config, path)
    if OmegaConf.is_config(selected):
        return OmegaConf.to_container(selected, resolve=True)
    else:
        return selected



# Requires !pip install azure-storage-blob==12.9.0

from azure.storage.blob import BlobServiceClient, BlobClient
from azure.core.exceptions import ServiceRequestError, ResourceNotFoundError, AzureError
from delphai_ml_utils.config import get_config

azure_blobs_connection_string = get_config("azure.blobs.connection_string")
azure_blobs_container = get_config("azure.blobs.folder_name")


def read_blob(blob_name: str) -> str:
    blob = BlobClient.from_connection_string(
        azure_blobs_connection_string, azure_blobs_container, blob_name
    )
    stream = blob.download_blob()
    data = stream.readall()
    content = data.decode("utf-8")
    return content


def blob_exist(key: str) -> bool:
    blob = BlobClient.from_connection_string(
        conn_str=azure_blobs_connection_string,
        container_name=azure_blobs_container,
        blob_name=key,
    )
    return blob.exists()



import os
from subprocess import STDOUT, check_call


def install_python38():
    check_call(
        ["sudo apt-get update -y"],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )
    check_call(
        ["sudo apt-get install python3.8"],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )
    check_call(
        [
            "sudo update-alternatives --install /usr/bin/python3 python3 /usr/bin/python3.8 4"
        ],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )
    check_call(
        ["sudo apt install python3-pip"],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )
    check_call(
        ["curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py"],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )
    check_call(
        ["python3 get-pip.py --force-reinstall"],
        shell=True,
        stdin=None,
        stdout=open(os.devnull, "wb"),
        stderr=STDOUT,
        executable="/bin/bash",
    )



import numpy as np
import torch


"""
For tokenizing long texts into the chunks (function tokenize_with_chunks)
tutorial used - https://www.kaggle.com/code/datajameson/long-text-sentiment-classification-bert-windowing/notebook

Example how to get predicted labels from probabilities of every text chunk 
in case of multilabel classification - function get_predictions
"""


def make_chunks(tokens: dict, chunksize=512, bos=0, eos=2, pad=1):
    """
    Make chunks of tokenized and encoded text
       input:
            tokens - dictionary, which is output of the tokenizer with no max_length limits
            chunksize - length of the chunk in tokens
            bos, eos, pad - special tokens encodings

       output: chunks of token ids and attention masks
    """

    # split into chunks of 510 tokens, we also convert to list (default is tuple which is immutable)
    input_id_chunks = list(tokens["input_ids"][0].split(chunksize - 2))
    mask_chunks = list(tokens["attention_mask"][0].split(chunksize - 2))

    # loop through each chunk
    for i in range(len(input_id_chunks)):
        # add BOS and EOS tokens to input IDs
        input_id_chunks[i] = torch.cat(
            [
                torch.tensor([bos], dtype=torch.long),
                input_id_chunks[i],
                torch.tensor([eos], dtype=torch.long),
            ]
        )
        # add attention for special tokens to the attention mask
        mask_chunks[i] = torch.cat(
            [torch.tensor([1]), mask_chunks[i], torch.tensor([1])]
        )
        # get required padding length
        pad_len = chunksize - input_id_chunks[i].shape[0]

        # check if tensor length satisfies required chunk size
        if pad_len > 0:
            # check if tensor length is more than 0, we must add padding
            input_id_chunks[i] = torch.cat(
                [input_id_chunks[i], torch.tensor([pad] * pad_len, dtype=torch.long)]
            )
            mask_chunks[i] = torch.cat(
                [mask_chunks[i], torch.tensor([0] * pad_len, dtype=torch.long)]
            )
    return torch.stack(input_id_chunks), torch.stack(mask_chunks)


def tokenize_with_chunks(tokenizer, texts, chunksize=512):
    """Tokenizing long texts into the chunks"""

    # IDs of special tokens
    if tokenizer.bos_token_id:
        bos = tokenizer.bos_token_id
    else:
        bos = tokenizer.cls_token_id  # bert-like models use CLS and SEP tokens
    if tokenizer.eos_token_id:
        eos = tokenizer.eos_token_id
    else:
        eos = tokenizer.sep_token_id
    pad = tokenizer.pad_token_id

    texts_indices = list()
    input_ids = list()
    attention_mask = list()

    for ind, text in enumerate(texts):
        # tokenize the text. No max_length limits, no special tokens
        tokens = tokenizer.encode_plus(
            text, add_special_tokens=False, return_tensors="pt"
        )
        # make chunks
        input_ids_, attention_mask_ = make_chunks(tokens, chunksize, bos, eos, pad)
        # save index of the text
        texts_indices.extend([ind] * input_ids_.shape[0])
        # append chunks to the list
        input_ids.append(input_ids_)
        attention_mask.append(attention_mask_)

    # stack token IDs and attention masks into tensors with the shape (number_of_chunks, chunksize)
    input_ids = torch.vstack(input_ids)
    attention_mask = torch.vstack(attention_mask)

    return input_ids, attention_mask, texts_indices


def get_text_labels(probs, idx2label, threshold):
    """
    Convert probabilities into labels for every chunk of the single news article's text.

    input:
        probs - 2D numpy array or list of lists with probabilties;
        idx2label - dictionary, index to label map;
        threshold - float, threshold for cutting the probabilities.

    output: list of lists of labels.
    """

    # make list of lists of booleans from probablities using threshold
    # we may consider the list of booleans as a one-hot vector (True=1 and False=0)
    bool_predictions = [[p >= threshold for p in prob] for prob in probs]

    # list with text predictions
    str_predictions = list()

    # loop over one-hot vectors in the list
    for pred in bool_predictions:
        # list with chunk predictions
        str_pred = list()
        # get indices where value in one-hot vector is not zero
        inds = np.argwhere(pred)
        # for every index (this is an idx of a label)...
        for ind in inds:
            # ... map it into a label and append string to the list of chunk predictions
            str_pred.append(idx2label[ind[0]])
        # store predicted labels of a chunk in a list
        str_predictions.append(str_pred)

    return str_predictions


def get_predictions(
    probabilities, idx2label, texts_indices, unique=False, threshold=0.5
):
    """
    Agregate predictions from probabilities into either list of lists of labels
    or list of unique labels.
    If argument `unique` is True only unique labels for the whole text are returned.
    If argument `unique` is False labels for every chunk of the text are returned.

    idx2label - dictionary, mapping of label's IDs into strings
    texts_indices - list, indicates index of the text in the dataset for every it's chunk
    """

    predictions = list()
    # get start indices and counts of chunks for texts
    _, indx, counts = np.unique(texts_indices, return_index=True, return_counts=True)

    end = 0

    for start, count in zip(indx, counts):
        end += count
        # slice probabilites so we get all chunks of the text together
        probs = probabilities[start:end]
        str_preds = get_text_labels(probs, idx2label, threshold)

        if unique:
            # flatten the list of lists
            str_preds = [lbl for lst in str_preds for lbl in lst]
            # get unique labels for the whole text
            str_preds = np.unique(str_preds).tolist()

        predictions.append(str_preds)

    return predictions
